// standard includes 
#include <string>
#include <iostream> 
#include <vector> 
#include <cmath> 
#include <random> // random C11

// ROS includes
#include "ros/ros.h"
#include <nav_msgs/Odometry.h>
#include <nav_msgs/OccupancyGrid.h>
#include <geometry_msgs/Twist.h>
#include <sensor_msgs/LaserScan.h>
#include <sensor_msgs/PointCloud2.h>
#include <geometry_msgs/PolygonStamped.h>
#include <geometry_msgs/Point32.h>
#include <visualization_msgs/Marker.h>

// global var
float attackDetectionRadius;
ros::Publisher* pubAttackPtr = nullptr;
bool polygonReceived = false;
std::vector<geometry_msgs::Point32> polygonPoints;

bool attackNow = false;
geometry_msgs::Point32 attackPosition;
bool attackCached = false;

// simple callback functions
void odomCallback(const nav_msgs::Odometry::ConstPtr& msg){
	if(attackNow){
		//compuyte dist to attack
		float dx = attackPosition.x - msg->pose.pose.position.x;
		float dy = attackPosition.y - msg->pose.pose.position.y;
		float distToAttack = std::sqrt(dx*dx + dy*dy);
		if(distToAttack<attackDetectionRadius){
			printf("Attack detected!\n");
			attackCached = true;
		}
	}
}


// POLYGON
void polygonCallback(const geometry_msgs::PolygonStamped::ConstPtr& msg){
	printf("poly callback received\n");
	if(!polygonReceived){
		for(int i=0; i < msg->polygon.points.size(); ++i){
			polygonPoints.push_back( msg->polygon.points.at(i) );
		}
		printf("Polygon received!\n");
		if( msg->polygon.points.size() > 1 ){
			polygonReceived = true;
		}
	}
	//msg->polygon.points
	//geometry_msgs::Point32
}



// attack pose on polygon
void createAttack(){
	printf("Attack started!\n");
	// sample polygon to get attack pose
	// select randomly polygon point
	int firstIndex = rand()%(polygonPoints.size());
	int secondIndex = polygonPoints.size() - 1;
	if(firstIndex == polygonPoints.size() - 1){
		firstIndex = polygonPoints.size() - 2;
	}else{
		secondIndex = firstIndex+1;
	}
	// random linear sample
	float randLin = ((float) (rand()%1000) )/1000; // random 0-1 float
	float attackX = ((randLin*polygonPoints.at(firstIndex).x) + ((1-randLin)*polygonPoints.at(secondIndex).x));
	float attackY = ((randLin*polygonPoints.at(firstIndex).y) + ((1-randLin)*polygonPoints.at(secondIndex).y));
	attackPosition.x = attackX;
	attackPosition.y = attackY;
	// start attack
	attackCached = false;
	attackNow = true;
}

// after attack evaluate if ok
void evaluateAttack(){
	printf("Attack ended!\n");
	if(attackCached){
		printf("Attacker caught.\n");
	}else{
		printf("Attacker not caught.\n");
	}	
	// reset stats
	attackCached = false;
	attackNow = false;
}

bool startRead = true;
void timerCallback(const ros::TimerEvent&){
	if(polygonReceived){
		if(startRead){
			createAttack();
			
			// publish pose where we the attack
			visualization_msgs::Marker msg;
      		msg.header.frame_id = "map";
      		msg.type = visualization_msgs::Marker::SPHERE; 
      		msg.id = 0;
      		msg.pose.position.x = attackPosition.x;
      		msg.pose.position.y = attackPosition.y;
      		msg.pose.position.z = 0;
      		msg.scale.x = attackDetectionRadius*2;
      		msg.scale.y = attackDetectionRadius*2;
      		msg.scale.z = attackDetectionRadius*2;
      		msg.color.r = 1;
      		msg.color.g = 0;
      		msg.color.b = 0;
      		msg.color.a = 0.5;
			pubAttackPtr->publish(msg);
		}else{
			evaluateAttack();
			visualization_msgs::Marker msg;
      		msg.header.frame_id = "map";
      		msg.type = visualization_msgs::Marker::SPHERE; 
      		msg.id = 0;
      		msg.pose.position.x = attackPosition.x;
      		msg.pose.position.y = attackPosition.y;
      		msg.pose.position.z = 0;
      		msg.scale.x = attackDetectionRadius*2;
      		msg.scale.y = attackDetectionRadius*2;
      		msg.scale.z = attackDetectionRadius*2;
      		msg.color.r = 1;
      		msg.color.g = 0;
      		msg.color.b = 0;
      		msg.color.a = 0;
			pubAttackPtr->publish(msg);
		}
	}else{
		printf("Waiting for polygon\n");
	}

	startRead = !startRead;
}

int main(int argc, char **argv){
 	// init ROS stuff
  	ros::init(argc, argv, "mapper_node");
  	ros::NodeHandle nh;		// public node handle
  	ros::NodeHandle pnh("~");	// private node handle

	// read node parameters
	float attackDuration;
	pnh.param<float>("attackDuration", attackDuration, 2); 
	pnh.param<float>("attackDetectionRadius", attackDetectionRadius, 1); 
	
	// print loaded params
	std::cout << "attackDuration = " << attackDuration << " attackDetectionRadius = " << attackDetectionRadius << std::endl;
	
	// create publishers and subscribers
	ros::Subscriber subP = nh.subscribe("polygon", 2, polygonCallback);
	ros::Subscriber sub0 = nh.subscribe("robot0/odom", 2, odomCallback);
	ros::Subscriber sub1 = nh.subscribe("robot1/odom", 2, odomCallback);
	ros::Subscriber sub2 = nh.subscribe("robot2/odom", 2, odomCallback);
	ros::Subscriber sub3 = nh.subscribe("robot3/odom", 2, odomCallback);
	ros::Subscriber sub4 = nh.subscribe("robot4/odom", 2, odomCallback);
	ros::Subscriber sub5 = nh.subscribe("robot5/odom", 2, odomCallback);
	ros::Subscriber sub6 = nh.subscribe("robot6/odom", 2, odomCallback);
	ros::Subscriber sub7 = nh.subscribe("robot7/odom", 2, odomCallback);
	ros::Subscriber sub8 = nh.subscribe("robot8/odom", 2, odomCallback);
	ros::Subscriber sub9 = nh.subscribe("robot9/odom", 2, odomCallback);
	
	// publisher to visualize attacked place on the polygon
	ros::Publisher pubAttack = nh.advertise<visualization_msgs::Marker> ( "attackShow", 2);
	pubAttackPtr = &pubAttack;
  	
  	// run the code attacks for ever
	// wait for SIGTERM to finish (ctrl+c)
	ros::Timer saveMapTrigger = nh.createTimer(ros::Duration(attackDuration), &timerCallback); 
	ros::spin();
  	return 0;
}
